from enum import Enum


class LogLevel(str, Enum):
    """Enum used to define the options for the log level in the cli"""

    INFO = 0
    DEBUG = 1


class FormatTypes(str, Enum):
    HAPIBD = "hapibd"
    ILASH = "ilash"
    GERMLINE = "germline"
    RAPID = "rapid"


class OverlapOptions(str, Enum):
    """Enum defining options for how the use wants to define
    overlap with the gene region of interest. Values can be contains or overlaps"""

    CONTAINS = "contains"
    OVERLAPS = "overlaps"



"""Module to faciliate the user in parsing the phenotype file by incorporating multiple ecodings, separators, and by handling multiple errors."""

import gzip
from logging import Logger
from pathlib import Path
from typing import Dict, List, Set, Tuple, TypeVar, Union

from log import CustomLogger

logger: Logger = CustomLogger.get_logger(__name__)

# creating a type annotation for the PhenotypeFileParser class
T = TypeVar("T", bound="PhenotypeFileParser")


class PhenotypeFileParser:
    """Parser used to read in the phenotype file. This will allow use to account for different delimiters in files as well as catch errors."""

    def __init__(self, filepath: Union[Path, str]) -> None:
        """Initialize the PhenotypeFileParser class.

        Parameters
        ----------
        filepath : Path | str
            filepath to the phenotype file that has case control status for individuals

        Raises
        ------
        FileNotFoundError
        """
        self.individuals: List[str] = []
        # we are going to make sure the filepath variable is a
        # PosixPath
        filepath = Path(filepath)

        # now we are going to try to create an attribute for
        # the input file
        if not filepath.exists():
            raise FileNotFoundError(f"The file {filepath} was not found")
        else:
            self.file: Path = filepath

    def __enter__(self) -> T:
        """Open the input file. Method determines the appropriate file type and open the file. Method is called automatically by the context manager.

        Raises
        ------
        OSError
            Raised if the program encounters an OSError while opening the file
        """
        suffix = self.file.suffix

        try:
            if suffix == ".gz":
                file = gzip.open(self.file, "rt")
            else:
                file = open(self.file, "r", encoding="utf-8")
        except OSError as e:
            logger.critical(e)
            logger.critical(
                f"Encountered the following error while trying to open the file: {self.file}"
            )
            raise OSError(
                f"Encountered the following error while trying to open the file: {self.file}"
            )

        self.opened_file = file

        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        """Close the resource when it is not longer being used. Used by the context manager."""
        self.opened_file.close()

    @staticmethod
    def _check_separator(line: str) -> str:
        """Determine what the separator for the file should be.

        Parameters
        ----------
        line : str
            this is the first or second line of the phenotype file
            depending on whether or not the file has a header.

        Returns
        -------
        str
            returns the separator as a string. At the moment, comma
            separated, tab separated, and pipe separated are supported.

        Raises
        ------
        ValueError
            raises a value error if the method doesn't identify a
            supported separator.
        """
        if len(line.split(",")) > 1:
            return ","
        elif len(line.split("\t")) > 1:
            return "\t"
        elif len(line.split("|")) > 1:
            return "|"
        else:
            raise ValueError(
                "The was no appropriate separator found for the file. Currently DRIVE supports: ',' or '\t' or '|'"
            )

    def _determine_status(
        self,
        line: list[str],
        phenotype_dict: Dict[str, Dict[str, Set[str]]],
        phenotype_indx: Dict[int, str],
    ) -> None:
        """Add the individual to the appropriate case/ control/exclusion list.

        Parameters
        ----------
        line: List[str]
            list of individuals status for each phenotype in the file

        phenotype_dict : Dict[str, Dict[str, Set[str]]]

            returns a tuple with three elements. The first element is a
            dictionary where the keys are phenotypes. Values are
            dictionaries where the keys are 'cases' or 'controls' or
            'excluded' and values are list of ids. The second element is
            a dictionary that maps the index of the phenotype in the
            header line to the phenotype name. The third element is the
            separator string
        """
        # pull out the correct grid id
        grid_id = line[0]
        # we need to keep track of the total list of grids used in the
        # analysis
        self.individuals.append(grid_id)

        # go through each value in the file
        for indx, value in enumerate(line[1:]):
            phenotype_mapping = phenotype_indx.get(indx)

            if value == "1" or value == "1.0":
                phenotype_dict[phenotype_mapping]["cases"].add(grid_id)
            elif value == "0" or value == "0.0":
                phenotype_dict[phenotype_mapping]["controls"].add(grid_id)
            # we are going to excluded on values na, n/a, -1, -1.
            # 0, "", " " to try to catch different values
            elif value.lower() in ["na", "n/a", "-1", "-1.0", " ", ""]:
                phenotype_dict[phenotype_mapping]["excluded"].add(grid_id)
            else:
                logger.warning(
                    f"The status for individual, {grid_id}, was not recognized. The status found in the file was {value} for phenotype {phenotype_mapping}. This individual will be added to the exclusion list but it is recommended that the user checks to ensure that this is not a typo in the phenotype file."
                )
                phenotype_dict[phenotype_mapping]["excluded"].append(grid_id)

    def _create_phenotype_dictionary(
        self,
        header_line: str,
    ) -> Tuple[Dict[str, Dict[str, Set[str]]], Dict[int, str], str]:
        """Function that will generate a dictionary where the keys are
        phenotypes and the values list of cases/exclusions/controls

        Parameters
        ----------
        header_line : str
            line from the phenotype file

        Returns
        -------
        Tuple[Dict[str, Dict[str, Set[str]]], Dict[int, str], str]
            returns a tuple with three elements. The first element is a
            dictionary where the keys are phenotypes. Values are
            dictionaries where the keys are 'cases' or 'controls' or
            'excluded' and values are list of ids. The second element is
            a dictionary that maps the index of the phenotype in the
            header line to the phenotype name. The third element is the
            separator string
        """

        # determining what the appropriate separator should be
        separator = PhenotypeFileParser._check_separator(header_line)

        logger.debug(f"Identified the separator, {separator}, in the file: {self.file}")

        # raise an error if there is no header line, otherwise determine all the phenotypes
        if not "grid" in header_line.lower() and not "grids" in header_line.lower():
            error_msg = "Expected the first line of the phenotype file to have a header line with a column called grid or grids."

            logger.critical(error_msg)

            raise ValueError(error_msg)
        else:
            split_line_phenotypes = header_line.strip("\n").split(separator)[1:]

        # creating a dictionary that will map an index position to a
        # phecode
        phenotype_indx = {}
        # creating a dictionary to keep track of who are cases and
        # controls
        phenotype_dict = {}

        # build each dictionary
        for indx, phenotype in enumerate(split_line_phenotypes):
            phenotype_indx[indx] = phenotype
            phenotype_dict[phenotype] = {
                "cases": set(),
                "controls": set(),
                "excluded": set(),
            }

        logger.debug(f"Phenotype index dictionary:\n {phenotype_indx}")
        logger.debug(f"Phenotype counts dictionary: \n {phenotype_dict}")
        return phenotype_dict, phenotype_indx, separator

    def parse_cases_and_controls(
        self,
    ) -> Tuple[Dict[str, Dict[str, Set[str]]], List[str]]:
        """Generate a list for cases, controls, and excluded individuals.

        Returns
        -------
        Tuple[Dict[str, Dict[str, List[str]]], List[str]]
            returns a tuple where the first element is a dictionary where
            the keys are the phenotypes and the values are dictionary of
            the case/controls/excluded individuals lists. The second
            element is a list of all grids from the file to be used as a
            cohort
        """

        (
            phenotype_dict,
            phenotype_indx_mappings,
            separator,
        ) = self._create_phenotype_dictionary(self.opened_file.readline())

        for line in self.opened_file:
            # we need to first check if there is a header row
            split_line = line.strip("\n").split(separator)

            self._determine_status(split_line, phenotype_dict, phenotype_indx_mappings)

        return phenotype_dict, self.individuals



from .choices import FormatTypes, LogLevel, OverlapOptions
from .data_container import Data, Data_Interface
from .generate_indices import FileIndices, create_indices
from .networks import Network, Network_Interface
from .types import Filter, Genes



from pathlib import Path


def check_input_exists(ibd_input_file: Path) -> Path:
    """Callback that will check that the input ibd file exists

    Parameters
    ----------
    ibd_input_file : Path
        Path object to the input ibd file. This file should be gzipped

    Returns
    -------
    Path
        returns the Path object if it exists

    Raises
    ------
    FileNotFoundError
        If the ibd input file does not exist then the program will
        immediately raise a FileNotFoundError
    """
    if ibd_input_file.exists():
        return ibd_input_file
    else:
        raise FileNotFoundError(f"The file, {ibd_input_file}, was not found")


def check_json_path(json_path: Path) -> Path:
    """Callback function that creates the json path string. If the user provides a value then it uses the user provided value else it creates the path to the default file
    Parameters
    ----------
    json_path : str
        path to the json config file or an empty string
    Returns
    -------
    str
        returns the string to the file
    """

    if json_path:
        return json_path
    else:
        root_dir = Path(__file__).parent.parent.parent.parent

        config_path = root_dir / "config.json"

        if not config_path.exists():
            raise FileNotFoundError(
                f"Expected the user to either pass a configuration file path or for the config.json file to be present in the program root directory at {config_path}."
            )

        return config_path



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'drive'}

packages = \
['drive',
 'drive.cluster',
 'drive.factory',
 'drive.filters',
 'drive.log',
 'drive.models',
 'drive.plugins',
 'drive.utilities',
 'drive.utilities.callbacks',
 'drive.utilities.parser',
 'factory']

package_data = \
{'': ['*']}

install_requires = \
['igraph>=0.10.4,<0.11.0',
 'numpy>=1.24.2,<2.0.0',
 'pandas>=1.5.3,<2.0.0',
 'scipy>=1.10.1,<2.0.0',
 'typer[all]>=0.7.0,<0.8.0']

entry_points = \
{'console_scripts': ['drive = drive.drive:app']}

setup_kwargs = {
    'name': 'drive-ibd',
    'version': '2.0.7',
    'description': 'cli tool to identify networks of individuals who share IBD segments overlapping loci of interest',
    'long_description': '# DRIVE:\n\nThis repository contains the source code for the tool DRIVE v1.0.0. Distant Relatedness for Identification and Variant Evaluation (DRIVE) is a novel approach to IBD-based genotype inference used to identify shared chromosomal segments in dense genetic arrays. DRIVE implemented a random walk algorithm that identifies clusters of individuals who pairwise share an IBD segment overlapping a locus of interest. This tool was developed in python by the Below Lab at Vanderbilt University. The documentation for how to use this tool can be found here [DRIVE documentation](https://belowlab.github.io/drive/)\n\n## Installing DRIVE:\nDRIVE is available on PYPI and can easily be installed using the following command:\n\n```bash\npip install drive-ibd\n```\n\nIf the user wishes to develop DRIVE or install the program from source then they can clone the repository. This process is described under the section called "Github Installation" in the documentation.\n\n### Reporting issues:\nIf you wish to report a bug or propose a feature you can find templates under the .github/ISSUE_TEMPLATE directory.',
    'author': 'James Baker',
    'author_email': 'james.baker@vanderbilt.edu',
    'maintainer': 'James Baker',
    'maintainer_email': 'james.baker@vanderbilt.edu',
    'url': 'https://belowlab.github.io/drive/',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<3.12',
}


setup(**setup_kwargs)
